/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CAENQdcController.h"
#include "drivers/global/DriversCommands.h"
#include "controllers/common/family/Families.h"
#include <boost/thread/thread.hpp>
#include <lstdpp128/lstdpp.h>

using namespace std;
using namespace boost;

namespace npp {

const string CAENQdcController::TYPE = "qdc_controller";

/*
 * Constructor
 */
CAENQdcController::CAENQdcController(const string& name) : CAENAdcController(name) {

	setFamily(family::ACQUISITION, family::SETTING);

	// Register properties
	groupChannel.init(this, NOSAVE, "group_channel");
	groupChannels.init(this, SAVE, "group_channels");
	displayGroup.init(this, SAVE, "display_group");
	chargeSensitivity.init(this, NOSAVE, "charge_sensitivity");
    inputSmoothing.init(this, NOSAVE, "input_smoothing");
	triggerHoldOff.init(this, NOSAVE, "trigger_holdoff");
	gateWidth.init(this, NOSAVE, "gate_width");
	preGate.init(this, NOSAVE, "pre_gate");
	baselineMean.init(this, NOSAVE, "base_mean");
	channelActive.init(this, NOSAVE, "channel_active");
	triggerThreshold.init(this, NOSAVE, "trigger_threshold");

	m_dgtzDriver.init(this, "qdc_driver");

	registerFunction(TYPE);
}

/*
 * Destructor
 */
CAENQdcController::~CAENQdcController() {

}

/*
 * postConfiguration
 */
void CAENQdcController::postConfiguration() {

	registerRefresher(recordLength, &CAENQdcController::refreshRecordLengthProperty, this);
	registerRefresher(dppAcqMode, &CAENQdcController::refreshDppAcqModeProperty, this);
	registerRefresher(displayGroup, &CAENQdcController::refreshDisplayGroupProperty, this);
	registerRefresher(triggerHoldOff, &CAENQdcController::refreshTriggerHoldOffProperty, this);
	registerRefresher(chargeSensitivity, &CAENQdcController::refreshChargeSensitivityProperty, this);
	registerRefresher(gateWidth, &CAENQdcController::refreshGateWidthProperty, this);
	registerRefresher(preGate, &CAENQdcController::refreshPreGateProperty, this);
	registerRefresher(baselineMean, &CAENQdcController::refreshBaselineMean, this);
	registerRefresher(dcOffset, &CAENQdcController::refreshDcOffsetProperty, this);
	registerRefresher(preTrigger, &CAENQdcController::refreshPreTriggerProperty, this);
	registerRefresher(pulsePolarity, &CAENQdcController::refreshPulsePolarityProperty, this);
	registerRefresher(channelActive, &CAENQdcController::refreshActivateChannelProperty, this);
	registerRefresher(triggerThreshold, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(groupChannel, &CAENQdcController::refreshGroupChannelProperty, this);

    triggerThreshold.resize(8);
	channelActive.resize(8);

	updateBoardProperties();

	refreshGroupChannelProperty(0);
}

void CAENQdcController::refreshRecordLengthProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->recordLength = aValue;
}

void CAENQdcController::refreshDppAcqModeProperty(int32 aValue)  throw (CannotSetValue) {

	CAENAdcController::refreshDppAcqModeProperty(aValue);
	m_dgtzDriver->dppAcqMode = aValue;
	m_dgtzDriver->acqMode = aValue;
	boost::thread td(boost::bind(&CAENQdcController::initBoard, this));
}

void CAENQdcController::refreshDisplayGroupProperty(int32 aGroup) throw (CannotSetValue) {

	for (int channel=0; channel<8; ++channel) {
		int32 aChannel = aGroup * 8 + channel;
		int32 trigTrsd = m_dgtzDriver->triggerThreshold.get(aChannel);
		int32 chActive = m_dgtzDriver->channelActive.get(aChannel);
		if (trigTrsd == 0)
			trigTrsd = 10;
		triggerThreshold.update(channel, trigTrsd);
		channelActive.update(channel, chActive);
	}

	float64 holdOff = m_dgtzDriver->triggerHoldOff.get(aGroup) * 16 / 1000.0;
	if (holdOff == 0)
		holdOff = 0.2;
	triggerHoldOff.update(holdOff);

	chargeSensitivity.update(m_dgtzDriver->chargeSensitivity());

	float64 gWidth = m_dgtzDriver->gateWidth.get(aGroup) * 16 / 1000.0;
	if (gWidth == 0.0)
		gWidth = 1.0;
	gateWidth.update(gWidth);

	float64 pGate = m_dgtzDriver->preGate.get(aGroup) * 16 / 1000.0;
	if (pGate == 0)
		pGate = 0.1;
	preGate.update(pGate);

	int32 baseAvrg = m_dgtzDriver->baselineMean();
	if (baseAvrg == 0)
		baseAvrg = 4;
	baselineMean.update(baseAvrg);

	int32 offSet = m_dgtzDriver->dcOffset.get(aGroup);
	if (offSet == 0)
		offSet = 10;
	dcOffset.update(offSet);

	int32 preTrig = m_dgtzDriver->preTrigger.get(aGroup);
	if (preTrig == 0)
		preTrig = 500;
	preTrigger.update(preTrig);

	pulsePolarity.update(m_dgtzDriver->pulsePolarity());

	groupChannel.update(0);
	refreshGroupChannelProperty(0);
}

void CAENQdcController::refreshTriggerHoldOffProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerHoldOff.set(displayGroup(), (int32)(aValue * 1000 / 16));
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshChargeSensitivityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->chargeSensitivity.set(aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshGateWidthProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->gateWidth.set(displayGroup(), (int32)(aValue * 1000 / 16));
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshPreGateProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preGate.set(displayGroup(), (int32)(aValue * 1000 / 16));
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshBaselineMean(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->baselineMean.set(aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshDcOffsetProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->dcOffset.set(displayGroup(), aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshPreTriggerProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preTrigger.set(displayGroup(), aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshPulsePolarityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->pulsePolarity.set(aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshActivateChannelProperty(int32 index, int32 aValue) throw (CannotSetValue) {

	int32 aChannel = displayGroup() * 8 + index;
	m_dgtzDriver->channelActive.set(aChannel, aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshTriggerThresholdProperty(int32 index, int32 aValue) throw (CannotSetValue) {

	int32 aChannel = displayGroup() * 8 + index;
	m_dgtzDriver->triggerThreshold.set(aChannel, aValue);
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::updateBoardProperties() {

	modelName.update(m_dgtzDriver->modelName());
	channels.update(m_dgtzDriver->channels());
	nbChannels.update(m_dgtzDriver->nbChannels());
	minRange = 0;
	maxRange = 7;
	dppAcqMode.update(m_dgtzDriver->dppAcqMode());
	cardType = lstdpp128::C1740_QDC;
}

void CAENQdcController::writeParam() {

	m_dgtzDriver.execute(caen_qdc::CAENQdcDriver::WRITE_PARAMETERS_COMMAND, false);
}

void CAENQdcController::initBoard() {

	m_dgtzDriver.execute(driver::INIT_COMMAND, false);
}

void CAENQdcController::refreshNbChannelsProperty(int32 aGroup) throw (CannotSetValue) {

}

void CAENQdcController::refreshDisplayChannelProperty(int32 aChannel) throw (CannotSetValue)  {

}

void CAENQdcController::refreshGroupChannelProperty(int32 aChannel) throw (CannotSetValue)  {

	displayChannel = displayGroup() * 8 + aChannel;
}

void CAENQdcController::refreshChannelActiveProperty(int32 aGroup) throw (CannotSetValue)  {

}

void CAENQdcController::initChannel(int32 aGroup) {

}

void CAENQdcController::clearData() {
	commandProgression = 0;
	m_dgtzDriver.execute(caen_qdc::CAENQdcDriver::CLEAR_COMMAND, true);
	commandProgression = 100;
}

}
