/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteNanokin.h"

#include <iostream>
#include <common/base/ServerProperties.h>
#include "controllers/common/family/Families.h"
#include "rapidjson/stringbuffer.h"
#include "rapidjson/prettywriter.h"

namespace nanokin {

using namespace std;
using namespace cameo;
using namespace rapidjson;

const std::string RemoteNanokin::TYPE = "remote_nanokin";

const std::string RemoteNanokin::NANOKIN_START_CMD = "vascokin-start";
const std::string RemoteNanokin::NANOKIN_STOP_CMD = "vascokin-stop";
const std::string RemoteNanokin::NANOKIN_FILETRANSFERT_CMD = "filetransfer";
const std::string RemoteNanokin::NANOKIN_STARTED_MSG = "Init Measure FPGA Done";
const std::string RemoteNanokin::START_ACTION = "start";
const std::string RemoteNanokin::END_ACTION = "end";

/*
 * Constructor
 */
RemoteNanokin::RemoteNanokin(const std::string& name) :
		ExperimentController(name), controller::Stoppable(this) {

	setFamily(family::SAMPLE_ENVIRONMENT);

	serverEndpoint.init(this, SAVE, "cameo_server");
	action.init(this, SAVE, "action");
	sopxFile.init(this, SAVE, "sopx_file");

	m_Title.init(this, "title");
	m_driver.init(this, "driver");

	m_initialized = false;
}

RemoteNanokin::RemoteNanokin(const RemoteNanokin& controller) : ExperimentController(controller), controller::Stoppable(this) {
	m_initialized = false;
}

/*
 * Destructor
 */
RemoteNanokin::~RemoteNanokin() {
}

void RemoteNanokin::postConfiguration() {

	if (m_server.get() == 0) {
		// Create a new server if it is not already created.
		// In case of a simulated server, avoid the real remote endpoint.
		if (serverEndpoint() == "" || m_driver->state() != REAL_DEVICE_CONTAINER) {
			m_server.reset(new Server(application::This::getServer().getEndpoint()));
		} else {
			m_server.reset(new Server(serverEndpoint()));
		}
	}
}

/*
 * start
 */
void RemoteNanokin::start() {

	commandProgression = 0;
	commandStatus.setRunning();

	if (action() == START_ACTION) {
		try {
			if (!m_server->isAvailable(10000)) {
				cout << "Nanokin cameo server is not available" << endl;
				commandStatus.setError();
				return;
			}

			cout << "Nanokin cameo server is connected to " << getName() << endl;

			//start args="START DefaultSOP.sopx  c:\\temp save.dat"
			vector<string> args;
			args.push_back(sopxFile());
			args.push_back("c:\\temp");
			args.push_back("save.dat");
			m_StartApplication = m_server->start(NANOKIN_START_CMD, args, cameo::OUTPUTSTREAM);

			m_StartOutputSocket = m_StartApplication->getOutputStreamSocket();
			cout << "m_StartOutputSocket = " << m_StartOutputSocket << endl;
			if (m_StartOutputSocket) {
				std::thread outputThread(std::bind(&RemoteNanokin::NanokinStartOutput, this));
				cout << "outputThread = " << outputThread.get_id() << endl;
				outputThread.join();
			}

		} catch (const std::exception & e) {
			// Currently an exception can occur during isAvailable.
			cout << "Unexpected exception during nanokin cameo connection: " << e.what() << endl;
		}
	}
	else if (action() == END_ACTION) {
		try {
			if (!m_server->isAvailable(10000)) {
				cout << "Nanokin cameo server is not available" << endl;
				return;
			}

			cout << "Nanokin cameo server is connected to " << getName() << endl;

			m_StopApplication = m_server->start(NANOKIN_STOP_CMD);

			if (m_StartApplication) {
				m_StartApplication->waitFor(application::STOPPED);
				if (m_StartOutputSocket) {
					m_StartOutputSocket->cancel();
				}
			}
		} catch (const std::exception & e) {
			// Currently an exception can occur during isAvailable.
			cout << "Unexpected exception during nanokin cameo connection: " << e.what() << endl;
		}

		try {
			if (!m_server->isAvailable(10000)) {
				cout << "Nanokin cameo server is not available" << endl;
				return;
			}

			cout << "Nanokin cameo server is connected to " << getName() << endl;

			cout << "m_FileTransfert connect" << endl;
			m_FileTransfert = m_server->connect(NANOKIN_FILETRANSFERT_CMD);

			if (m_FileTransfert->exists()) {
				// The application exists from a previous server session
	//			m_FileTransfert->stop();
	//			application::State state = m_FileTransfert->waitFor();
			}
			else {
				cout << "m_FileTransfert start" << endl;
				m_FileTransfert = m_server->start(NANOKIN_FILETRANSFERT_CMD);
			}

			m_FileRequester = application::Requester::create(*m_FileTransfert, "file-transfer");

			StringBuffer s;
			PrettyWriter<StringBuffer> writer(s);
			writer.StartObject();
			writer.Key("operation");
			writer.String("read");
			writer.Key("type");
			writer.String("binary");
			writer.Key("path");
			writer.String("c:\\temp\\save.dat");
			writer.EndObject();

			cout << "send message" << endl;
			m_FileRequester->send(s.GetString());

			string response;
			cout << "wait message" << endl;
			m_FileRequester->receiveBinary(response);
			if (response.empty() == false) {
				ostringstream datafile;
				datafile << common::ServerProperties::getInstance()->getNomadDataPath() << boost::format("%06d") % m_Title->numor() << ".dat";
				ofstream file(datafile.str().c_str(), fstream::binary);
				file.write(response.data(), response.length());
				file.close();
				cout << "Save : " << datafile.str() << " " << response.length() << endl;

				StringBuffer s1;
				PrettyWriter<StringBuffer> writer1(s1);
				writer1.StartObject();
				writer1.Key("operation");
				writer1.String("delete");
				writer1.Key("type");
				writer1.String("binary");
				writer1.Key("path");
				writer1.String("c:\\temp\\save.dat");
				writer1.EndObject();

				cout << "send message" << endl;
				m_FileRequester->send(s1.GetString());
				string response1;
				cout << "wait message" << endl;
				m_FileRequester->receiveBinary(response1);
			}
			m_FileRequester.release();
		} catch (const std::exception & e) {
			// Currently an exception can occur during isAvailable.
			cout << "Unexpected exception during nanokin cameo connection: " << e.what() << endl;
		}
	}
	commandProgression = 100;
	commandStatus.setIdle();
}

/*
 * stop
 */
void RemoteNanokin::stop() {
	if (m_StartOutputSocket) {
		m_StartOutputSocket->cancel();
	}
	if (m_FileRequester) {
		m_FileRequester->cancel();
	}
}

/*
 * NanokinStartOutput
 */
void RemoteNanokin::NanokinStartOutput() {
	while (true) {
		Output output;
		if (m_StartOutputSocket->receive(output)) {
			cout << output.getMessage() << endl;
			if (output.getMessage().find(NANOKIN_STARTED_MSG) != string::npos) {
				return;
			}
		}
		else {
			return;
		}
	}
}

}
