/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "GenericCamDriver.h"

#include <math.h>
#include "PerfectGenericCamDriver.h"
#include "GenericCamDef.h"
#include "GenericCamState.h"
#include "RealGenericCamDriver.h"
#include "SimulatedGenericCamDriver.h"


using namespace std;

namespace genericcam {

const std::string GenericCamDriver::TYPE = "genericcam";
const std::string GenericCamDriver:: MODE[4] = {"Standard", "Average","Summation","Accumulation"};

/*
 * Constructor
 */
GenericCamDriver::GenericCamDriver(const string& name):
						acquisition::NoTofAcquisitionCommon(name),acquisition::KineticAcquisitionCommon(name),
						acquisition::TimeGateCommon(name),
						acquisition::DetectorCommon(name)
						{

	registerStates(new RealGenericCamDriver(this), new PerfectGenericCamDriver(this), new SimulatedGenericCamDriver(this));

	driver::SocketTcp::init(name);

	nbCols.init(this, SAVE, "nbCols");
	nbRows.init(this, SAVE, "nbRows");

	height.init(this, SAVE, "height");
	width.init(this, SAVE, "width");
	acqmode.init(this, SAVE, "acqmode");
	gain.init(this, SAVE, "gain");
	nbFrame.init(this, SAVE, "nbFrame");
	bpixel.init(this, SAVE, "bpixel");
	fileSimu.init(this, SAVE, "file_simu");
	indexSimu.init(this, SAVE, "index_simu");
	fileDirectory.init(this, SAVE, "fileDirectory");
	fileName.init(this, SAVE, "fileName");
	numor.init(this, SAVE, "numor");
	header.init(this, NOSAVE, "header");
	aTimouOut.init(this, SAVE, "camtimeout");

	initCommand(driver::INIT_COMMAND);
	initCommand(driver::READ_COMMAND);
	initCommand(driver::SYNCHRONIZE_READ_COMMAND);
	initCommand(driver::START_COMMAND);
	initCommand(driver::RESUME_COMMAND);
	initCommand(driver::PAUSE_COMMAND);
	initCommand(driver::STOP_COMMAND);
	initCommand(driver::STATUS_COMMAND);
	initCommand(driver::READ_INFOS_COMMAND);
	initCommand(DetectorCommon::CLEAR_COMMAND);
	initCommand(DetectorCommon::REGROUP_COMMAND);
	initCommand(AcquisitionCommon::WRITE_PARAMETERS_COMMAND);
	initCommand(driver::INIT_COMMAND);

	// Init functions
	registerFunction(NONE_FUNCTION);

	deviceType = LEAF_DEVICE_TYPE_DEVICE_CONTAINER;
	m_DoSynchroniseRead = false;
	m_SynchroniseReadActivated = false;
	m_SynchroniseReadTerminated = false;
	m_DataShort = NULL;

	registerObserverCommand(COMMAND_STATUS_DEVICE_CONTAINER, 50);
}

/*
 * Destructor
 */
GenericCamDriver::~GenericCamDriver() {
}

/*
 * refreshDataSizeProperty
 */
void GenericCamDriver::refreshDataSizeProperty(int32 value) throw (CannotSetValue) {
	if (m_DataShort != NULL) {
		delete[] m_DataShort;
	}
	m_DataShort = new uint16[value];
	if (m_DataShort == NULL) {
		sendErrorEvent(NOTENOUGH_MEMORY_ERROR, "Can't create data table");
		throw CannotSetValue();
	}
	DetectorCommon::refreshDataSizeProperty(value);
}

/*
 * refreshModeProperty
 */
void GenericCamDriver::refreshModeProperty(int32 value) throw (CannotSetValue) {
	if (value != acquisition::GateCommon::TIME_MODE) {
		throw CannotSetValue();
	}
}

/*
 * execute
 */
void GenericCamDriver::execute(const std::string& aCommand) {

	if ((aCommand != driver::PAUSE_COMMAND) && (aCommand != driver::STOP_COMMAND) && (aCommand != driver::STATUS_COMMAND)
			&& (aCommand != driver::READ_COMMAND)) {
		commandProgression = PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER;
	}
	GenericCamState* currentState = dynamic_cast<GenericCamState*>(getCurrentState());
	// Check command
	cout << "GenericCamDriver::stop() " << aCommand<< endl;


	if (aCommand == driver::INIT_COMMAND) {
		// Init command
		currentState->init();
	} else if (aCommand == driver::READ_COMMAND) {
		// Read command
		currentState->read();
	} else if (aCommand == driver::SYNCHRONIZE_READ_COMMAND) {
		// Read command
		if (m_DoSynchroniseRead == true) {
			m_SynchroniseReadActivated = true;
			m_SynchroniseReadTerminated = false;
			currentState->synchroniseRead();
			m_DoSynchroniseRead = false;
		}
		else {
			m_SynchroniseReadActivated = false;
			m_SynchroniseReadTerminated = false;
		}
	} else if (aCommand == driver::START_COMMAND) {
		// Start command
		startActivated = true;
		m_DoSynchroniseRead = true;
		currentState->start();
	} else if (aCommand == driver::RESUME_COMMAND) {
		// Continue command
		currentState->resume();
	} else if (aCommand == driver::PAUSE_COMMAND) {
		// Stop command
		currentState->pause();
	} else if (aCommand == driver::STOP_COMMAND) {
		// Stop command
		currentState->stop();
		startActivated = false;
	} else if (aCommand == driver::STATUS_COMMAND) {
		// Status command
		currentState->readStatus();
	} else if (aCommand == driver::READ_INFOS_COMMAND) {
		// Info command
		currentState->readInfos();
	} else if (aCommand == DetectorCommon::CLEAR_COMMAND) {
		// Move command
		sum = 0;
		currentState->clear();
	} else if (aCommand == DetectorCommon::REGROUP_COMMAND) {
//		currentState->regroup();
	} else if (aCommand == AcquisitionCommon::WRITE_PARAMETERS_COMMAND) {
		// Write command
		currentState->writeParam();
	} else {
		//Error bad command
		//Todo , normaly it will be detect it device action driver
	}
}

/*
 * calculateProgression
 */
void GenericCamDriver::calculateProgression() {
	int32 progression = 0;
	float64 setpoint = 0.0;
	float64 actual = 0.0;

	setpoint = time.setpoint();
	actual = time();

	if (setpoint == 0.0) {
		commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
	} else {
		progression = (int32) (actual / setpoint * PROGRESSION_RATIO_DEVICE_CONTAINER);
		// Be sure that progression > 0%
		if (progression < PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER)
			progression = 0;
		// Be sure that only Phased status put 100%
		if (progression >= PROGRESSION_END_DEVICE_CONTAINER)
			progression = PROGRESSION_END_DEVICE_CONTAINER - 1;
		commandProgression = progression;
	}
}

/*
 * GenericCamDriver
 */
void GenericCamDriver::computeGateStatus(int32 value) {
	int32 valueout = 0;
	if (value == COUNTING) {
		valueout |= GateCommon::RUNNING_STATUS;
	} else if (value == END) {
		valueout |= GateCommon::END_STATUS;
	}
	status.update(value);
	gateStatus.update(valueout);
}

/*
 * GenericCamDriver
 */
void GenericCamDriver::computeDetectorStatus(int32 value) {
	int32 valueout = 0;
	detectorStatus.update(valueout);

}

/*
 * GenericCamDriver
 */
void GenericCamDriver::computeAcquisitionStatus(int32 value) {
	int32 valueout = 0;
	acquisitionStatus.update(valueout);
}

}

