/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <drivers/ioxos/cadaracheacq/CadaracheAcqDef.h>
#include <drivers/ioxos/cadaracheacq/CadaracheAcqDriver.h>
#include <drivers/ioxos/cadaracheacq/CadaracheAcqState.h>
#include <drivers/ioxos/cadaracheacq/PerfectCadaracheAcqDriver.h>
#include <drivers/ioxos/cadaracheacq/RealCadaracheAcqDriver.h>
#include <drivers/ioxos/cadaracheacq/SimulatedCadaracheAcqDriver.h>

using namespace std;
using namespace acquisition;

namespace cadaracheacq {

const string CadaracheAcqDriver::TYPE = "CadaracheAcq";

/*
 * Constructor
 */
CadaracheAcqDriver::CadaracheAcqDriver(const string& aName) :
		KineticGateCommon(aName),
		KineticAcquisitionCommon(aName),
		DetectorCommon(aName)
{

	driver::Vme::init(aName);

	m_SynchroniseReadActivated = false;
	m_SynchroniseReadTerminated = false;
	m_StopActivated = false;

	// Init command list
	initCommand(driver::INIT_COMMAND);
	initCommand(driver::STATUS_COMMAND);
	initCommand(driver::START_COMMAND);
	initCommand(driver::READ_COMMAND);
	initCommand(driver::SYNCHRONIZE_READ_COMMAND);
	initCommand(driver::STOP_COMMAND);
	initCommand(driver::PAUSE_COMMAND);
	initCommand(driver::RESUME_COMMAND);
	initCommand(driver::READ_INFOS_COMMAND);
	initCommand(AcquisitionCommon::WRITE_PARAMETERS_COMMAND);
	initCommand(AcquisitionCommon::READ_PARAMETERS_COMMAND);
	initCommand(DetectorCommon::REGROUP_COMMAND);
	initCommand(DetectorCommon::CLEAR_COMMAND);

	registerStates(new RealCadaracheAcqDriver(this), new PerfectCadaracheAcqDriver(this), new PerfectCadaracheAcqDriver(this));

	// Init functions
	registerFunction(NONE_FUNCTION);

	deviceType = LEAF_DEVICE_TYPE_DEVICE_CONTAINER;
	addressType.update(Vme::A24_ACCESS);

	// Register the Spy and Observer commands necessary to do the updates.
	registerObserverCommand(COMMAND_READ_DEVICE_CONTAINER, 50);
	registerObserverCommand(COMMAND_STATUS_DEVICE_CONTAINER, 2);
}

/*
 * Destructor
 */
CadaracheAcqDriver::~CadaracheAcqDriver() {
}

/*
 * execute
 */
void CadaracheAcqDriver::execute(const string& aCommand) {

	if ((aCommand != driver::PAUSE_COMMAND) && (aCommand != driver::STOP_COMMAND) && (aCommand != driver::STATUS_COMMAND)
			&& (aCommand != driver::READ_COMMAND)) {
		commandProgression = PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER;
	}
	CadaracheAcqState* currentState = dynamic_cast<CadaracheAcqState*>(getCurrentState());
	// Check command
	if (aCommand == driver::INIT_COMMAND) {
		// Init command
		currentState->init();
	} else if (aCommand == driver::STATUS_COMMAND) {
		// Status command
		currentState->readStatus();
	} else if (aCommand == driver::START_COMMAND) {
		m_StopActivated = false;
		// Start command
		startActivated = true;
		currentState->start();
	} else if (aCommand == driver::READ_COMMAND) {
		// Read command
		currentState->read();
	} else if (aCommand == driver::SYNCHRONIZE_READ_COMMAND) {
		// Read command
		m_SynchroniseReadActivated = true;
		m_SynchroniseReadTerminated = false;
		currentState->synchroniseRead();
	} else if (aCommand == driver::STOP_COMMAND) {
		m_StopActivated = true;
		// Stop command
		currentState->stop();
		startActivated = false;
	} else if (aCommand == driver::PAUSE_COMMAND) {
		// Stop command
		currentState->pause();
	} else if (aCommand == driver::RESUME_COMMAND) {
		// Continue command
		currentState->resume();
	} else if (aCommand == driver::READ_INFOS_COMMAND) {
		// Info command
		currentState->readInfos();
	} else if (aCommand == AcquisitionCommon::WRITE_PARAMETERS_COMMAND) {
		// Write param command
		currentState->writeParam();
	} else if (aCommand == AcquisitionCommon::READ_PARAMETERS_COMMAND) {
		// Read param command
		currentState->readParam();
	} else if (aCommand == DetectorCommon::REGROUP_COMMAND) {
		// do nothing
	} else if (aCommand == DetectorCommon::CLEAR_COMMAND) {
		currentState->clear();
	}
}

/*
 * calculateProgression
 */
void CadaracheAcqDriver::calculateProgression() {
	int32 progression = 0;


	float64 setpoint = nbRepaint.setpoint() * nbSlicesAcq();
	float64 actual = (nbRepaint() - 1) * nbSlicesAcq() + actualSlice();

	progression = (int32) (actual / setpoint * PROGRESSION_RATIO_DEVICE_CONTAINER);
	// Be sure that progression > 0%
	if (progression < PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER) {
		progression = 0;
	}
	// Be sure that only Phased status put 100%
	if (progression >= PROGRESSION_END_DEVICE_CONTAINER) {
		progression = PROGRESSION_END_DEVICE_CONTAINER - 1;
	}
	commandProgression = progression;
}

/*
 * computeGateStatus
 */
void CadaracheAcqDriver::computeGateStatus(int32 value) {
	int32 valueout = 0;
	if (((value & GATE_EXT_STATUS) == 0) && ((value & FUNCTION_START_ACTIVE_STATUS) != 0)) {
		valueout |= GateCommon::WAITFORSTART_STATUS;
	} else if (value & RUNNING_STATUS) {
		valueout |= GateCommon::RUNNING_STATUS;
	}
	gateStatus.update(valueout);
}

/*
 * computeAcquisitionStatus
 */
void CadaracheAcqDriver::computeAcquisitionStatus(int32 value) {
	int32 valueout = 0;
	acquisitionStatus.update(valueout);
}

}
